@@ -8,42 +8,45 @@ Functions for bootstrapping elements
 """
 
 from __future__ import annotations
-from meerschaum.utils.typing import Union, Any, Sequence, SuccessTuple, Optional, Tuple, List
+
+import meerschaum as mrsm
+from meerschaum.utils.typing import Union, Any, SuccessTuple, Optional, Tuple, List
+
 
 def bootstrap(
-        action: Optional[List[str]] = None,
-        **kw: Any
-    ) -> SuccessTuple:
+    action: Optional[List[str]] = None,
+    **kw: Any
+) -> SuccessTuple:
     """
-    Bootstrap an element (pipes, connectors, config).
-    
+    Launch an interactive wizard to bootstrap pipes or connectors.
+
     Example:
         `bootstrap pipes`
 
     """
-    from meerschaum.utils.misc import choose_subaction
-    if action is None:
-        action = []
+    from meerschaum.actions import choose_subaction
     options = {
         'pipes'      : _bootstrap_pipes,
-        'config'     : _bootstrap_config,
         'connectors' : _bootstrap_connectors,
+        'plugins'    : _bootstrap_plugins,
+        'jobs'       : _bootstrap_jobs,
     }
     return choose_subaction(action, options, **kw)
 
+
 def _bootstrap_pipes(
-        action: Optional[List[str]] = None,
-        connector_keys: Optional[List[str]] = None,
-        metric_keys: Optional[List[str]] = None,
-        location_keys: Optional[List[Optional[str]]] = None,
-        yes: bool = False,
-        force: bool = False,
-        noask: bool = False,
-        debug: bool = False,
-        mrsm_instance: Optional[str] = None,
-        shell: bool = False,
-        **kw: Any
-    ) -> SuccessTuple:
+    action: Optional[List[str]] = None,
+    connector_keys: Optional[List[str]] = None,
+    metric_keys: Optional[List[str]] = None,
+    location_keys: Optional[List[Optional[str]]] = None,
+    yes: bool = False,
+    force: bool = False,
+    noask: bool = False,
+    debug: bool = False,
+    mrsm_instance: Optional[str] = None,
+    shell: bool = False,
+    **kw: Any
+) -> SuccessTuple:
     """
     Create a new pipe.
     If no keys are provided, guide the user through the steps required.
@@ -51,10 +54,9 @@ def _bootstrap_pipes(
     """
     from meerschaum import get_pipes
     from meerschaum.config import get_config
-    from meerschaum.utils.warnings import info, warn, error
-    from meerschaum.utils.debug import dprint
+    from meerschaum.utils.warnings import info, warn
     from meerschaum.utils.prompt import yes_no, prompt, choose
-    from meerschaum.connectors.parse import is_valid_connector_keys, parse_instance_keys
+    from meerschaum.connectors.parse import parse_instance_keys
     from meerschaum.utils.misc import get_connector_labels
     from meerschaum.utils.formatting._shell import clear_screen
 
@@ -99,17 +101,19 @@ def _bootstrap_pipes(
         ### Get the connector.
         new_label = 'New'
         info(
-            f"To create a pipe without explicitly using a connector, "
+            "To create a pipe without explicitly using a connector, "
             + "use the `register pipes` command.\n"
         )
         try:
             ck = choose(
-                f"Where are the data coming from?\n\n" +
-                f"    Please type the keys of a connector from below,\n" +
-                f"    or enter '{new_label}' to register a new connector.\n\n" +
-                f" {get_config('formatting', 'emoji', 'connector')} Connector:\n",
+                (
+                    "Where are the data coming from?\n\n" +
+                    f"    Please type the keys of a connector or enter '{new_label}'\n" +
+                    "    to register a new connector.\n\n" +
+                    f" {get_config('formatting', 'emoji', 'connector')} Connector:"
+                ),
                 get_connector_labels() + [new_label],
-                numeric = False
+                numeric = False,
             )
         except KeyboardInterrupt:
             return abort_tuple
@@ -128,7 +132,7 @@ def _bootstrap_pipes(
                     break
                 elif isinstance(tup[0], bool) and not tup[0]:
                     return abort_tuple
-                warn(f"Please register a new connector or press CTRL+C to cancel.", stack=False)
+                warn("Please register a new connector or press CTRL+C to cancel.", stack=False)
         connector_keys = [ck]
 
         ### Get the metric.
@@ -137,9 +141,9 @@ def _bootstrap_pipes(
                 clear_screen(debug=debug)
             try:
                 mk = prompt(
-                    f"What kind of data is this?\n\n" +
-                    f"    The metric is the label for the contents of the pipe.\n" +
-                    f"    For example, 'weather' might be a metric for weather station data.\n\n" +
+                    "What kind of data is this?\n\n" +
+                    "    The metric is the label for the contents of the pipe.\n" +
+                    "    For example, 'weather' might be a metric for weather station data.\n\n" +
                     f" {get_config('formatting', 'emoji', 'metric')} Metric:"
                 )
             except KeyboardInterrupt:
@@ -160,7 +164,7 @@ def _bootstrap_pipes(
                 + f"    For example, you could create Pipe('{ck}', '{mk}', 'home') and\n"
                 + f"    Pipe('{ck}', '{mk}', 'work'), which would share a connector\n"
                 + "    and metric, but may come from different tables.\n\n"
-                + "    In most cases. you can omit the location.\n\n"
+                + "    In most cases, you can omit the location.\n\n"
                 + f" {get_config('formatting', 'emoji', 'location')} Location (Empty to omit):"
             )
         except KeyboardInterrupt:
@@ -219,33 +223,39 @@ def _bootstrap_pipes(
 
     return (successes > 0), msg
 
+
 def _bootstrap_connectors(
-        action : Optional[List[str]] = None,
-        connector_keys : Optional[List[str]] = None,
-        yes : bool = False,
-        force : bool = False,
-        noask : bool = False,
-        debug : bool = False,
-        return_keys : bool = False,
-        **kw : Any
-    ) -> Union[SuccessTuple, Tuple[str, str]]:
+    action: Optional[List[str]] = None,
+    connector_keys: Optional[List[str]] = None,
+    yes: bool = False,
+    force: bool = False,
+    noask: bool = False,
+    debug: bool = False,
+    return_keys: bool = False,
+    **kw: Any
+) -> Union[SuccessTuple, Tuple[str, str]]:
     """
     Prompt the user for the details necessary to create a Connector.
-
     """
-    from meerschaum.connectors.parse import is_valid_connector_keys
-    from meerschaum.connectors import connectors, get_connector, types, custom_types
+    from meerschaum.connectors import (
+        connectors,
+        get_connector,
+        types,
+        custom_types,
+        _load_builtin_custom_connectors,
+    )
     from meerschaum.utils.prompt import prompt, yes_no, choose
     from meerschaum.config import get_config
     from meerschaum.config._edit import write_config
     from meerschaum.utils.formatting import pprint
     from meerschaum.utils.formatting._shell import clear_screen
-    from meerschaum.connectors import attributes as connector_attributes
+    from meerschaum.config._default import CONNECTOR_ATTRIBUTES
     from meerschaum.utils.warnings import warn, info
     from meerschaum.utils.misc import is_int
 
     abort_tuple = False, "No connectors bootstrapped."
     _clear = get_config('shell', 'clear_screen', patch=True)
+    _load_builtin_custom_connectors()
 
     if action is None:
         action = []
@@ -259,11 +269,11 @@ def _bootstrap_connectors(
         _type = choose(
             (
                 'Please choose a connector type.\n'
-                + 'For more information on connectors, '
-                + 'please visit https://meerschaum.io/reference/connectors'
+                + '    See https://meerschaum.io/reference/connectors '
+                + 'for documentation on connectors.\n'
             ),
-            sorted(list(connectors)),
-            default = 'sql'
+            sorted([k for k in connectors if k != 'plugin']),
+            default='sql',
         )
     except KeyboardInterrupt:
         return abort_tuple
@@ -288,21 +298,30 @@ def _bootstrap_connectors(
             warn(f"Connector '{_type}:{_label}' already exists.", stack=False)
             overwrite = yes_no(
                 f"Do you want to overwrite connector '{_type}:{_label}'?",
-                default = 'n',
-                yes = yes,
-                noask = noask,
+                default='n',
+                yes=yes,
+                noask=noask,
             )
             if not overwrite and not force:
-                return False, f"No changes made to connector configuration."
-                break
+                return False, "No changes made to connector configuration."
+            break
         elif _label == "":
-            warn(f"Please enter a label.", stack=False)
+            warn("Please enter a label.", stack=False)
         else:
             break
 
     cls = types.get(_type)
     cls_required_attrs = getattr(cls, 'REQUIRED_ATTRIBUTES', [])
-    type_attributes = connector_attributes.get(_type, {'required': cls_required_attrs})
+    cls_optional_attrs = getattr(cls, 'OPTIONAL_ATTRIBUTES', [])
+    cls_default_attrs = getattr(cls, 'DEFAULT_ATTRIBUTES', {})
+    type_attributes = CONNECTOR_ATTRIBUTES.get(
+        _type,
+        {
+            'required': cls_required_attrs,
+            'optional': cls_optional_attrs,
+            'default': cls_default_attrs,
+        }
+    )
 
     new_attributes = {}
     if 'flavors' in type_attributes:
@@ -311,18 +330,20 @@ def _bootstrap_connectors(
                 f"Flavor for connector '{_type}:{_label}':",
                 sorted(list(type_attributes['flavors'])),
                 default = (
-                    'timescaledb'
-                    if 'timescaledb' in type_attributes['flavors']
+                    'postgresql'
+                    if 'postgresql' in type_attributes['flavors']
                     else None
                 )
             )
         except KeyboardInterrupt:
             return abort_tuple
         new_attributes['flavor'] = flavor
-        required = sorted(list(connector_attributes[_type]['flavors'][flavor]['requirements']))
+        required = sorted(list(CONNECTOR_ATTRIBUTES[_type]['flavors'][flavor]['requirements']))
+        optional = sorted(list(CONNECTOR_ATTRIBUTES[_type]['flavors'][flavor].get('optional', {})))
         default = type_attributes['flavors'][flavor].get('defaults', {})
     else:
         required = sorted(list(type_attributes.get('required', {})))
+        optional = sorted(list(type_attributes.get('optional', {})))
         default = type_attributes.get('default', {})
     info(
         f"Please answer the following questions to configure the new connector '{_type}:{_label}'."
@@ -330,13 +351,22 @@ def _bootstrap_connectors(
     )
     for r in required:
         try:
-            val = prompt(f"Value for {r}:")
+            default_val = str(default.get(r)) if r in default else None
+            val = prompt(f"Value for {r}:", default=default_val)
         except KeyboardInterrupt:
             continue
         if is_int(val):
             val = int(val)
         new_attributes[r] = val
 
+    for o in optional:
+        try:
+            val = prompt(f"Value for {o} (optional; empty to omit):")
+        except KeyboardInterrupt:
+            continue
+        if val:
+            new_attributes[o] = val
+
     for k, v in default.items():
         ### skip already configured attributes, (e.g. flavor or from required)
         if k in new_attributes:
@@ -378,44 +408,144 @@ def _bootstrap_connectors(
     if _type not in meerschaum_config['connectors']:
         meerschaum_config['connectors'][_type] = {}
     meerschaum_config['connectors'][_type][_label] = new_attributes
-    write_config({'meerschaum' : meerschaum_config}, debug=debug)
+    write_config({'meerschaum': meerschaum_config}, debug=debug)
     if return_keys:
         return _type, _label
     return True, "Success"
 
-def _bootstrap_config(
-        action : Optional[List[str]] = None,
-        yes : bool = False,
-        force : bool = False,
-        debug : bool = False,
-        **kw : Any
-    ) -> SuccessTuple:
-    """
-    Delete and regenerate the default Meerschaum configuration.
 
+def _bootstrap_plugins(
+    action: Optional[List[str]] = None,
+    debug: bool = False,
+    **kwargs: Any
+) -> SuccessTuple:
     """
-    from meerschaum.config._edit import write_default_config, write_config
-    from meerschaum.config._default import default_config
+    Launch an interactive wizard to guide the user to creating a new plugin.
+    """
+    from meerschaum.utils.prompt import prompt, yes_no
+    from meerschaum.plugins.bootstrap import bootstrap_plugin
+    from meerschaum.config.paths import PLUGINS_DIR_PATHS
 
+    for dir_path in PLUGINS_DIR_PATHS:
+        if not dir_path.exists():
+            create_dir_path = yes_no(f"Create path '{dir_path}'?", default='y')
+            if create_dir_path:
+                dir_path.mkdir(parents=True, exist_ok=True)
+            else:
+                return False, "Create missing plugins paths before bootstrapping."
+
+    if not action:
+        action = [prompt("Enter the name of your new plugin:")]
+
+    for plugin_name in action:
+        bootstrap_success, bootstrap_msg = bootstrap_plugin(plugin_name, debug=debug)
+        if not bootstrap_success:
+            return bootstrap_success, bootstrap_msg
+
+    return True, "Success"
+
+
+def _bootstrap_jobs(
+    action: Optional[List[str]] = None,
+    executor_keys: Optional[str] = None,
+    debug: bool = False,
+    **kwargs: Any
+) -> SuccessTuple:
+    """
+    Launch an interactive wizard to create new jobs.
+    """
+    import shlex
+    from meerschaum.utils.prompt import prompt, yes_no
     from meerschaum.actions import actions
-    if not actions['delete'](['config'], debug=debug, yes=yes, force=force, **kw)[0]:
-        return False, "Failed to delete configuration files."
+    from meerschaum.utils.formatting import print_options, make_header
+    from meerschaum.utils.formatting._shell import clear_screen
+    from meerschaum.utils.warnings import info
+    from meerschaum._internal.arguments import (
+        split_pipeline_sysargs,
+        split_chained_sysargs,
+    )
+    from meerschaum.utils.misc import items_str
+    from meerschaum._internal.shell.ShellCompleter import ShellCompleter
 
-    if not write_config(default_config, debug=debug, **kw):
-        return (False, "Failed to write default configuration.")
+    if not action:
+        action = [prompt("What is the name of the job you'd like to create?")]
 
-    if not write_default_config(debug=debug, **kw):
-        return (False, "Failed to write default configuration.")
+    for name in action:
+        clear_screen(debug=debug)
+        job = mrsm.Job(name, executor_keys=executor_keys)
+        if job.exists():
+            edit_success, edit_msg = actions['edit'](['job', name], **kwargs)
+            if not edit_success:
+                return edit_success, edit_msg
+            continue
 
-    from meerschaum.config.stack import write_stack
-    if not write_stack(debug=debug):
-        return False, "Failed to write stack."
+        info(
+            f"Editing arguments for job '{name}'.\n"
+            "    Press [Esc + Enter] to submit, [CTRL + C] to exit.\n\n"
+            "    Tip: join multiple actions with `+`, add pipeline arguments with `:`.\n"
+            "    https://meerschaum.io/reference/actions/#chaining-actions\n"
+        )
+        try:
+            new_sysargs_str = prompt(
+                "",
+                multiline=True,
+                icon=False,
+                completer=ShellCompleter(),
+            )
+        except KeyboardInterrupt:
+            return True, "Nothing was changed."
 
-    return (True, "Successfully bootstrapped configuration files.")
+        new_sysargs = shlex.split(new_sysargs_str)
+        new_sysargs, pipeline_args = split_pipeline_sysargs(new_sysargs)
+        chained_sysargs = split_chained_sysargs(new_sysargs)
+
+        clear_screen(debug=debug)
+        if len(chained_sysargs) > 1:
+            print_options(
+                [
+                    shlex.join(step_sysargs)
+                    for step_sysargs in chained_sysargs
+                ],
+                header=f"Steps in Job '{name}':",
+                number_options=True,
+                **kwargs
+            )
+        else:
+            print('\n' + make_header(f"Action for Job '{name}':"))
+            print(shlex.join(new_sysargs))
+
+        if pipeline_args:
+            print('\n' + make_header("Pipeline Arguments:"))
+            print(shlex.join(pipeline_args))
+            print()
+
+        if not yes_no(
+            (
+                f"Are you sure you want to create job '{name}' with the above arguments?\n"
+                + "    The job will be started if you continue."
+            ),
+            default='n',
+            **kwargs
+        ):
+            return True, "Nothing was changed."
+
+        new_job = mrsm.Job(name, new_sysargs_str, executor_keys=executor_keys)
+        start_success, start_msg = new_job.start()
+        if not start_success:
+            return start_success, start_msg
+
+    msg = (
+        "Successfully bootstrapped job"
+        + ('s' if len(action) != 1 else '')
+        + ' '
+        + items_str(action)
+        + '.'
+    )
+    return True, msg
 
 
 ### NOTE: This must be the final statement of the module.
 ###       Any subactions added below these lines will not
 ###       be added to the `help` docstring.
-from meerschaum.utils.misc import choices_docstring as _choices_docstring
+from meerschaum.actions import choices_docstring as _choices_docstring
 bootstrap.__doc__ += _choices_docstring('bootstrap')
