@@ -9,19 +9,19 @@ Functions for deleting elements.
 from __future__ import annotations
 from meerschaum.utils.typing import Any, SuccessTuple, Union, Optional, List
 
+
 def delete(
-        action: Optional[List[str]] = None,
-        **kw: Any
-    ) -> SuccessTuple:
+    action: Optional[List[str]] = None,
+    **kw: Any
+) -> SuccessTuple:
     """
     Delete an element.
 
     Command:
-        `delete {config, pipes, plugins, users, connectors, jobs}`
+        `delete {config, pipes, plugins, users, connectors, jobs, venvs}`
 
     """
-    from meerschaum.utils.misc import choose_subaction
-    from meerschaum.utils.debug import dprint
+    from meerschaum.actions import choose_subaction
     options = {
         'config'     : _delete_config, 
         'pipes'      : _delete_pipes,
@@ -29,27 +29,27 @@ def delete(
         'users'      : _delete_users,
         'connectors' : _delete_connectors,
         'jobs'       : _delete_jobs,
+        'venvs'      : _delete_venvs,
     }
     return choose_subaction(action, options, **kw)
 
+
 def _complete_delete(
-        action: Optional[List[str]] = None,
-        **kw: Any
-    ) -> List[str]:
+    action: Optional[List[str]] = None,
+    **kw: Any
+) -> List[str]:
     """
     Override the default Meerschaum `complete_` function.
-
     """
-    from meerschaum.actions.start import _complete_start_jobs
     from meerschaum.actions.edit import _complete_edit_config
     if action is None:
         action = []
     options = {
         'connector': _complete_delete_connectors,
         'connectors': _complete_delete_connectors,
-        'config' : _complete_edit_config,
-        'job' : _complete_start_jobs,
-        'jobs' : _complete_start_jobs,
+        'config': _complete_edit_config,
+        'job': _complete_delete_jobs,
+        'jobs': _complete_delete_jobs,
     }
 
     if (
@@ -63,28 +63,41 @@ def _complete_delete(
     from meerschaum._internal.shell import default_action_completer
     return default_action_completer(action=(['delete'] + action), **kw)
 
+
 def _delete_pipes(
-        debug: bool = False,
-        yes: bool = False,
-        force: bool = False,
-        noask: bool = False,
-        **kw: Any
-    ) -> SuccessTuple:
+    debug: bool = False,
+    yes: bool = False,
+    force: bool = False,
+    noask: bool = False,
+    **kw: Any
+) -> SuccessTuple:
     """
     Drop pipes and delete their registrations.
-
     """
     from meerschaum import get_pipes
     from meerschaum.utils.prompt import yes_no
-    from meerschaum.utils.formatting import pprint, highlight_pipes
+    from meerschaum.utils.formatting import highlight_pipes
     from meerschaum.utils.warnings import warn
+    from meerschaum.actions import actions
     pipes = get_pipes(as_list=True, debug=debug, **kw)
     if len(pipes) == 0:
         return False, "No pipes to delete."
+
+    _ = kw.pop('action', None)
+    _ = actions['drop'](
+        ['pipes'],
+        yes = yes,
+        force = force,
+        noask = noask,
+        debug = debug,
+        **kw
+    )
+
     question = "Are you sure you want to delete these pipes? This can't be undone!\n"
     for p in pipes:
         question += f"    - {p}" + "\n"
     question = highlight_pipes(question)
+
     answer = force
     if force:
         answer = True
@@ -97,7 +110,7 @@ def _delete_pipes(
     success_dict = {}
 
     for p in pipes:
-        success_tuple = p.delete(debug=debug)
+        success_tuple = p.delete(drop=False, debug=debug)
         success_dict[p] = success_tuple[1]
         if success_tuple[0]:
             successes += 1
@@ -112,21 +125,22 @@ def _delete_pipes(
     
     return successes > 0, msg
 
+
 def _delete_config(
-        action: Optional[List[str]] = None,
-        yes: bool = False,
-        force: bool = False,
-        noask: bool = False,
-        debug: bool = False,
-        **kw: Any
-    ) -> SuccessTuple:
+    action: Optional[List[str]] = None,
+    yes: bool = False,
+    force: bool = False,
+    noask: bool = False,
+    debug: bool = False,
+    **kw: Any
+) -> SuccessTuple:
     """
     Delete configuration files.
-
     """
-    import os, shutil
+    import os
+    import shutil
     from meerschaum.utils.prompt import yes_no
-    from meerschaum.config._paths import CONFIG_DIR_PATH, STACK_COMPOSE_PATH, DEFAULT_CONFIG_DIR_PATH
+    from meerschaum.config._paths import STACK_COMPOSE_PATH, DEFAULT_CONFIG_DIR_PATH
     from meerschaum.config._read_config import get_possible_keys, get_keyfile_path
     from meerschaum.utils.debug import dprint
     paths = [p for p in [STACK_COMPOSE_PATH, DEFAULT_CONFIG_DIR_PATH] if p.exists()]
@@ -143,9 +157,11 @@ def _delete_config(
     else:
         sep = '\n' + '  - '
         answer = yes_no(
-            f"Are you sure you want to delete the following configuration files?" +
+            "Are you sure you want to delete the following configuration files?" +
             f"{sep + sep.join([str(p) for p in paths])}\n",
-            default='n', noask=noask, yes=yes
+            default='n',
+            noask=noask,
+            yes=yes,
         )
 
     if answer or force:
@@ -162,18 +178,18 @@ def _delete_config(
     
     return success, msg
 
+
 def _delete_plugins(
-        action: Optional[List[str]] = None,
-        repository: Optional[str] = None,
-        yes: bool = False,
-        force: bool = False,
-        noask: bool = False,
-        debug: bool = False,
-        **kw: Any
-    ) -> SuccessTuple:
+    action: Optional[List[str]] = None,
+    repository: Optional[str] = None,
+    yes: bool = False,
+    force: bool = False,
+    noask: bool = False,
+    debug: bool = False,
+    **kw: Any
+) -> SuccessTuple:
     """
     Delete plugins from a Meerschaum repository.
-
     """
     from meerschaum.utils.warnings import info
     from meerschaum.plugins import reload_plugins
@@ -192,7 +208,7 @@ def _delete_plugins(
     ) if not force else True
 
     if not answer:
-        return False, f"No plugins deleted."
+        return False, "No plugins deleted."
 
     successes = {}
     for name in action:
@@ -204,27 +220,24 @@ def _delete_plugins(
     reload_plugins(debug=debug)
     return True, "Success"
 
+
 def _delete_users(
-        action: Optional[List[str]] = None,
-        mrsm_instance: Optional[str] = None,
-        yes: bool = False,
-        force: bool = False,
-        noask: bool = False,
-        shell: bool = False,
-        debug: bool = False,
-        **kw
-    ) -> SuccessTuple:
+    action: Optional[List[str]] = None,
+    mrsm_instance: Optional[str] = None,
+    yes: bool = False,
+    force: bool = False,
+    noask: bool = False,
+    shell: bool = False,
+    debug: bool = False,
+    **kw
+) -> SuccessTuple:
     """
     Delete users from a Meerschaum instance. Adequate permissions are required.
-
     """
-    from meerschaum import get_connector
     from meerschaum.connectors.parse import parse_instance_keys
-    from meerschaum.utils.prompt import yes_no, prompt
-    from meerschaum.utils.debug import dprint
-    from meerschaum.utils.warnings import warn, error, info
+    from meerschaum.utils.prompt import yes_no
+    from meerschaum.utils.warnings import info
     from meerschaum.core import User
-    from meerschaum.connectors.api import APIConnector
     from meerschaum.utils.formatting import print_tuple
     instance_connector = parse_instance_keys(mrsm_instance)
 
@@ -278,24 +291,25 @@ def _delete_users(
     )
     return True, msg
 
+
 def _delete_connectors(
-        action: Optional[List[str]] = None,
-        connector_keys: Optional[List[str]] = None,
-        yes: bool = False,
-        force: bool = False,
-        noask: bool = False,
-        debug: bool = False,
-        **kw: Any
-    ) -> SuccessTuple:
+    action: Optional[List[str]] = None,
+    connector_keys: Optional[List[str]] = None,
+    yes: bool = False,
+    force: bool = False,
+    noask: bool = False,
+    debug: bool = False,
+    **kw: Any
+) -> SuccessTuple:
     """
     Delete configured connectors.
     
     Example:
         `delete connectors sql:test`
-
     """
-    import os, pathlib
-    from meerschaum.utils.prompt import yes_no, prompt
+    import os
+    import pathlib
+    from meerschaum.utils.prompt import yes_no
     from meerschaum.connectors.parse import parse_connector_keys
     from meerschaum.config import _config
     from meerschaum.config._edit import write_config
@@ -315,7 +329,7 @@ def _delete_connectors(
     for ck in _keys:
         try:
             conn = parse_connector_keys(ck, debug=debug)
-        except Exception as e:
+        except Exception:
             warn(f"Could not parse connector '{ck}'. Skipping...", stack=False)
             continue
 
@@ -335,7 +349,7 @@ def _delete_connectors(
     for c in to_delete:
         try:
             ### Remove database files.
-            if c.flavor in ('sqlite', 'duckdb'):
+            if c.flavor in ('sqlite', 'duckdb', 'geopackage'):
                 if ':memory:' not in c.database and pathlib.Path(c.database).exists():
                     if force or yes_no(
                         f"Detected '{c.flavor}' database '{c.database}'. "
@@ -343,26 +357,27 @@ def _delete_connectors(
                     ):
                         try:
                             os.remove(c.database)
-                        except Exception as e:
+                        except Exception:
                             warn(
                                 "Failed to delete database file for connector "
                                 + f"'{c}'. Ignoring...", stack=False
                             )
-        except Exception as e:
+        except Exception:
             pass
         try:
             del cf['meerschaum']['connectors'][c.type][c.label]
-        except Exception as e:
+        except Exception:
             warn(f"Failed to delete connector '{c}' from configuration. Skipping...", stack=False)
 
     write_config(cf, debug=debug)
     return True, "Success"
 
+
 def _complete_delete_connectors(
-        action: Optional[List[str]] = None,
-        line: str = '',
-        **kw: Any
-    ) -> List[str]:
+    action: Optional[List[str]] = None,
+    line: str = '',
+    **kw: Any
+) -> List[str]:
     from meerschaum.config import get_config
     from meerschaum.utils.misc import get_connector_labels
     types = list(get_config('meerschaum', 'connectors').keys())
@@ -372,23 +387,27 @@ def _complete_delete_connectors(
         search_term = action[-1]
     return get_connector_labels(*types, search_term=search_term)
 
+
 def _delete_jobs(
-        action : Optional[List[str]] = None,
-        noask : bool = False,
-        nopretty : bool = False,
-        force : bool = False,
-        yes : bool = False,
-        debug : bool = False,
-        **kw
-    ) -> SuccessTuple:
+    action: Optional[List[str]] = None,
+    executor_keys: Optional[str] = None,
+    noask: bool = False,
+    nopretty: bool = False,
+    force: bool = False,
+    yes: bool = False,
+    debug: bool = False,
+    **kw
+) -> SuccessTuple:
     """
     Remove a job's log files and delete the job's ID.
     
     If the job is running, ask to kill the job first.
-
     """
-    from meerschaum.utils.daemon import (
-        Daemon, get_running_daemons, get_stopped_daemons, get_filtered_daemons,
+    from meerschaum.jobs import (
+        get_running_jobs,
+        get_stopped_jobs,
+        get_filtered_jobs,
+        get_paused_jobs,
     )
     from meerschaum.utils.prompt import yes_no
     from meerschaum.utils.formatting._jobs import pprint_jobs
@@ -396,46 +415,54 @@ def _delete_jobs(
     from meerschaum.utils.warnings import warn
     from meerschaum.utils.misc import items_str
     from meerschaum.actions import actions
-    daemons = get_filtered_daemons(action, warn=(not nopretty))
-    if not daemons:
-        return False, "No jobs to delete."
+
+    jobs = get_filtered_jobs(executor_keys, action, debug=debug)
+    if not jobs:
+        return True, "No jobs to delete; nothing to do."
+
     _delete_all_jobs = False
     if not action:
         if not force:
-            pprint_jobs(daemons)
+            pprint_jobs(jobs)
             if not yes_no(
                 "Delete all jobs? This cannot be undone!",
                 noask=noask, yes=yes, default='n'
             ):
                 return False, "No jobs were deleted."
+
             _delete_all_jobs = True
-    _running_daemons = get_running_daemons(daemons)
-    _stopped_daemons = get_stopped_daemons(daemons, _running_daemons)
-    _to_delete = _stopped_daemons
-    if _running_daemons:
+
+    _running_jobs = get_running_jobs(executor_keys, jobs, debug=debug)
+    _paused_jobs = get_paused_jobs(executor_keys, jobs, debug=debug)
+    _stopped_jobs = get_stopped_jobs(executor_keys, jobs, debug=debug)
+    _to_delete = _stopped_jobs
+
+    to_stop_jobs =  {**_running_jobs, **_paused_jobs}
+    if to_stop_jobs:
         clear_screen(debug=debug)
         if not force:
-            pprint_jobs(_running_daemons, nopretty=nopretty)
+            pprint_jobs(to_stop_jobs, nopretty=nopretty)
         if force or yes_no(
-            "The above jobs are still running. Stop these jobs?",
+            "Stop these jobs?",
             default='n', yes=yes, noask=noask
         ):
             actions['stop'](
-                action = (['jobs'] + [d.daemon_id for d in _running_daemons]),
-                nopretty = nopretty,
-                yes = yes,
-                force = force,
-                noask = noask,
-                debug = debug,
+                action=(['jobs'] + [_name for _name in to_stop_jobs]),
+                executor_keys=executor_keys,
+                nopretty=nopretty,
+                yes=yes,
+                force=force,
+                noask=noask,
+                debug=debug,
+                **kw
             )
-            #  import time
-            #  time.sleep(1)
             ### Ensure the running jobs are dead.
-            if get_running_daemons(daemons):
+            if get_running_jobs(executor_keys, jobs, debug=debug):
                 return False, (
-                    f"Failed to kill running jobs. Please stop these jobs before deleting."
+                    "Failed to kill running jobs. Please stop these jobs before deleting."
                 )
-            _to_delete += _running_daemons
+            _to_delete.update(to_stop_jobs)
+
         ### User decided not to kill running jobs.
         else:
             pass
@@ -447,28 +474,146 @@ def _delete_jobs(
         pprint_jobs(_to_delete, nopretty=nopretty)
         if not yes_no(
             "Are you sure you want to delete these jobs?",
-            yes=yes, noask=noask, default='y',
+            yes=yes, noask=noask, default='n',
         ):
             return False, "No jobs were deleted."
 
     _deleted = []
-    for d in _to_delete:
-        d.cleanup()
-        if d.path.exists() and not nopretty:
-            warn(f"Failed to delete job '{d.daemon_id}'.", stack=False)
+    for name, job in _to_delete.items():
+        delete_success, delete_msg = job.delete()
+        if not delete_success:
+            warn(f"Failed to delete job '{name}'.", stack=False)
             continue
-        _deleted.append(d)
+        _deleted.append(name)
 
     return (
         len(_deleted) > 0,
         ("Deleted job" + ("s" if len(_deleted) != 1 else '')
-            + f" {items_str([d.daemon_id for d in _deleted])}."),
+            + f" {items_str([_name for _name in _deleted])}."),
     )
 
 
+def _complete_delete_jobs(
+    action: Optional[List[str]] = None,
+    executor_keys: Optional[str] = None,
+    line: str = '',
+    _get_job_method: Union[str, List[str], None] = None,
+    **kw
+) -> List[str]:
+    from meerschaum._internal.shell.Shell import shell_attrs
+    from meerschaum.jobs import (
+        get_jobs,
+        get_filtered_jobs,
+        get_restart_jobs,
+        get_stopped_jobs,
+        get_paused_jobs,
+        get_running_jobs,
+        get_executor_keys_from_context,
+    )
+    from meerschaum.utils.misc import remove_ansi
+    from meerschaum.connectors.parse import parse_executor_keys
+
+    executor_keys = (
+        executor_keys
+        or remove_ansi(
+            shell_attrs.get('executor_keys', get_executor_keys_from_context())
+        )
+    )
+
+    if (
+        executor_keys != 'systemd'
+        and parse_executor_keys(executor_keys, construct=False) is None
+    ):
+        return []
+
+    jobs = get_jobs(executor_keys, include_hidden=False)
+    if _get_job_method:
+        method_keys = [_get_job_method] if isinstance(_get_job_method, str) else _get_job_method
+        method_jobs = {}
+        for method_key in method_keys:
+            method_func = locals()[f'get_{method_key}_jobs']
+            method_jobs.update(method_func(jobs=jobs))
+        jobs = method_jobs
+
+    if not action:
+        return list(jobs)
+
+    possibilities = []
+    _line_end = line.split(' ')[-1]
+    for name in jobs:
+        if name in action:
+            continue
+        if _line_end == '':
+            possibilities.append(name)
+            continue
+        if name.startswith(action[-1]):
+            possibilities.append(name)
+    return possibilities
+
+
+def _delete_venvs(
+    action: Optional[List[str]] = None,
+    yes: bool = False,
+    force: bool = False,
+    **kwargs: Any
+) -> SuccessTuple:
+    """
+    Remove virtual environments.
+    Specify which venvs to remove, or remove everything at once.
+    """
+    import os
+    import shutil
+    import pathlib
+    from meerschaum.config.paths import VIRTENV_RESOURCES_PATH
+    from meerschaum.utils.venv import venv_exists
+    from meerschaum.utils.prompt import yes_no
+    from meerschaum.utils.misc import print_options
+
+    venvs_to_skip = ['mrsm']
+    venvs = [
+        _venv
+        for _venv in action or os.listdir(VIRTENV_RESOURCES_PATH)
+        if venv_exists(_venv)
+        and _venv not in venvs_to_skip
+    ]
+
+    if not venvs:
+        msg = "No venvs to delete."
+        if action:
+            return False, msg
+        return True, msg
+
+    print_options(
+        venvs,
+        header = 'Venvs to Delete:',
+        **kwargs
+    )
+    confirm_delete = yes_no(
+        (
+            "Remove the above venv" + ('s' if len(venvs) != 1 else '') + "?\n    "
+            + "Run `mrsm upgrade packages` and `mrsm install required` to reinstall dependencies.\n"
+        ),
+        yes = yes,
+        default = 'n',
+        force = force,
+    )
+    if not confirm_delete:
+        return True, "Nothing was deleted."
+
+    for venv in venvs:
+        venv_path = pathlib.Path(VIRTENV_RESOURCES_PATH / venv)
+        try:
+            shutil.rmtree(venv_path)
+        except Exception as e:
+            error_msg = f"Failed to remove '{venv_path}':\n{e}"
+            return False, error_msg
+
+    msg = f"Removed {len(venvs)} venv" + ('s' if len(venvs) != 1 else '') + '.'
+    return True, msg
+
 
 ### NOTE: This must be the final statement of the module.
 ###       Any subactions added below these lines will not
 ###       be added to the `help` docstring.
-from meerschaum.utils.misc import choices_docstring as _choices_docstring
+from meerschaum.actions import choices_docstring as _choices_docstring
 delete.__doc__ += _choices_docstring('delete')
