@@ -9,14 +9,25 @@ and if interactive, print the welcome message.
 
 from __future__ import annotations
 
-import os, shutil, sys, pathlib, copy
+import os
+import shutil
+import sys
+import copy
+import contextlib
+
 from meerschaum.utils.typing import Any, Dict, Optional, Union
 from meerschaum.utils.threading import RLock
-from meerschaum.utils.warnings import warn
 
 from meerschaum.config._version import __version__
 from meerschaum.config._edit import edit_config, write_config
-from meerschaum.config.static import STATIC_CONFIG
+from meerschaum.config._read_config import (
+    search_and_substitute_config,
+    revert_symlinks_config,
+    get_possible_keys,
+    get_keyfile_path,
+    read_config,
+)
+from meerschaum._internal.static import STATIC_CONFIG
 
 from meerschaum.config._paths import (
     PERMANENT_PATCH_DIR_PATH,
@@ -24,38 +35,66 @@ from meerschaum.config._paths import (
     DEFAULT_CONFIG_DIR_PATH,
 )
 from meerschaum.config._patch import (
-    #  permanent_patch_config,
-    #  patch_config,
     apply_patch_to_config,
 )
-__all__ = ('get_plugin_config', 'write_plugin_config', 'get_config', 'write_config', 'set_config',)
+__all__ = (
+    'get_plugin_config',
+    'write_plugin_config',
+    'get_config',
+    'write_config',
+    'edit_config',
+    'set_config',
+    'replace_config',
+    'search_and_substitute_config',
+    'revert_symlinks_config',
+    'get_possible_keys',
+    'get_keyfile_path',
+    'apply_patch_to_config',
+    'read_config',
+    'paths',
+    'STATIC_CONFIG',
+)
 __pdoc__ = {'static': False, 'resources': False, 'stack': False, }
 _locks = {'config': RLock()}
 
 ### apply config preprocessing (e.g. main to meta)
 config = {}
+_backup_config = None
+_allow_write_missing: bool = True
+
+
 def _config(
-        *keys: str, reload: bool = False, substitute: bool = True,
-        sync_files: bool = True, write_missing: bool = True,
-    ) -> Dict[str, Any]:
+    *keys: str,
+    reload: bool = False,
+    substitute: bool = True,
+    sync_files: bool = True,
+    allow_replaced: bool = True,
+    write_missing: bool = True,
+) -> Dict[str, Any]:
     """
     Read and process the configuration file.
     """
-    global config
+    global config, _backup_config
+
     if config is None or reload:
         with _locks['config']:
             config = {}
+
     if keys and keys[0] not in config:
         from meerschaum.config._sync import sync_files as _sync_files
         key_config = read_config(
-            keys=[keys[0]],
-            substitute=substitute,
-            write_missing=write_missing,
+            keys = [keys[0]],
+            substitute = substitute,
+            write_missing = write_missing and _allow_write_missing,
         )
         if keys[0] in key_config:
             config[keys[0]] = key_config[keys[0]]
-            if sync_files:
+            if sync_files and _allow_write_missing:
                 _sync_files(keys=[keys[0] if keys else None])
+
+    if not allow_replaced:
+        return _backup_config if _backup_config is not None else config
+
     return config
 
 
@@ -73,15 +112,15 @@ def set_config(cf: Dict[str, Any]) -> Dict[str, Any]:
 
 
 def get_config(
-        *keys: str,
-        patch: bool = True,
-        substitute: bool = True,
-        sync_files: bool = True,
-        write_missing: bool = True,
-        as_tuple: bool = False,
-        warn: bool = True,
-        debug: bool = False
-    ) -> Any:
+    *keys: str,
+    patch: bool = True,
+    substitute: bool = True,
+    sync_files: bool = True,
+    write_missing: bool = True,
+    as_tuple: bool = False,
+    warn: bool = True,
+    debug: bool = False
+) -> Any:
     """
     Return the Meerschaum configuration dictionary.
     If positional arguments are provided, index by the keys.
@@ -131,7 +170,11 @@ def get_config(
         dprint(f"Indexing keys: {keys}", color=False)
 
     if len(keys) == 0:
-        _rc = _config(substitute=substitute, sync_files=sync_files, write_missing=write_missing)
+        _rc = _config(
+            substitute=substitute,
+            sync_files=sync_files,
+            write_missing=(write_missing and _allow_write_missing),
+        )
         if as_tuple:
             return True, _rc 
         return _rc
@@ -147,18 +190,18 @@ def get_config(
     ):
         try:
             _subbed = search_and_substitute_config({keys[0]: config[keys[0]]})
-        except Exception as e:
+        except Exception:
             import traceback
             traceback.print_exc()
+            _subbed = {keys[0]: config[keys[0]]}
+
         config[keys[0]] = _subbed[keys[0]]
         if symlinks_key in _subbed:
             if symlinks_key not in config:
                 config[symlinks_key] = {}
-            if keys[0] not in config[symlinks_key]:
-                config[symlinks_key][keys[0]] = {}
-            config[symlinks_key][keys[0]] = apply_patch_to_config(
-                _subbed,
-                config[symlinks_key][keys[0]]
+            config[symlinks_key] = apply_patch_to_config(
+                _subbed.get(symlinks_key, {}),
+                config.get(symlinks_key, {}),
             )
 
     from meerschaum.config._sync import sync_files as _sync_files
@@ -187,7 +230,7 @@ def get_config(
         for k in keys:
             try:
                 c = c[k]
-            except Exception as e:
+            except Exception:
                 invalid_keys = True
                 break
         if invalid_keys:
@@ -202,7 +245,7 @@ def get_config(
             for k in keys:
                 try:
                     _c = _c[k]
-                except Exception as e:
+                except Exception:
                     in_default = False
             if in_default:
                 c = _c
@@ -213,7 +256,7 @@ def get_config(
                     if warn:
                         from meerschaum.utils.warnings import warn as _warn
                         _warn(warning_msg, stacklevel=3, color=False)
-                except Exception as e:
+                except Exception:
                     if warn:
                         print(warning_msg)
                 if as_tuple:
@@ -232,7 +275,6 @@ def get_config(
                 )
             )
             if patch and keys[0] != symlinks_key:
-                #  print("Updating configuration, please wait...")
                 if write_missing:
                     write_config(config, debug=debug)
 
@@ -241,31 +283,43 @@ def get_config(
     return c
 
 
-def get_plugin_config(*keys : str, **kw : Any) -> Optional[Any]:
+def get_plugin_config(
+    *keys: str,
+    warn: bool = False,
+    **kw: Any
+) -> Optional[Any]:
     """
     This may only be called from within a Meerschaum plugin.
     See `meerschaum.config.get_config` for arguments.
     """
-    from meerschaum.utils.warnings import warn, error
+    from meerschaum.utils.warnings import error
     from meerschaum.plugins import _get_parent_plugin
     parent_plugin_name = _get_parent_plugin(2)
     if parent_plugin_name is None:
-        error(f"You may only call `get_plugin_config()` from within a Meerschaum plugin.")
-    return get_config(*(['plugins', parent_plugin_name] + list(keys)), **kw)
+        error(
+            "You may only call `get_plugin_config()` "
+            "from within a Meerschaum plugin."
+        )
+
+    return get_config(
+        *(['plugins', parent_plugin_name] + list(keys)),
+        warn=warn,
+        **kw
+    )
 
 
 def write_plugin_config(
-        config_dict: Dict[str, Any],
-        **kw : Any
-    ):
+    config_dict: Dict[str, Any],
+    **kw: Any
+):
     """
     Write a plugin's configuration dictionary.
     """
-    from meerschaum.utils.warnings import warn, error
+    from meerschaum.utils.warnings import error
     from meerschaum.plugins import _get_parent_plugin
     parent_plugin_name = _get_parent_plugin(2)
     if parent_plugin_name is None:
-        error(f"You may only call `get_plugin_config()` from within a Meerschaum plugin.")
+        error("You may only call `get_plugin_config()` from within a Meerschaum plugin.")
     plugins_cf = get_config('plugins', warn=False)
     if plugins_cf is None:
         plugins_cf = {}
@@ -274,13 +328,44 @@ def write_plugin_config(
     return write_config(cf, **kw)
 
 
+@contextlib.contextmanager
+def replace_config(config_: Union[Dict[str, Any], None]):
+    """
+    Temporarily override the Meerschaum config dictionary.
+
+    Parameters
+    ----------
+    config_: Dict[str, Any]
+        The new config dictionary to temporarily replace the canonical `config`.
+    """
+    if config_ is None:
+        try:
+            yield
+        finally:
+            return
+
+    global _backup_config, _allow_write_missing
+
+    _backup_config = _config()
+    _allow_write_missing = False
+    set_config(config_)
+
+    try:
+        yield
+    finally:
+        set_config(_backup_config)
+        _allow_write_missing = True
+
 ### This need to be below get_config to avoid a circular import.
 from meerschaum.config._read_config import read_config
 
 ### If environment variable MRSM_CONFIG or MRSM_PATCH is set, patch config before anything else.
-from meerschaum.config._environment import apply_environment_patches, apply_environment_uris
-apply_environment_uris()
-apply_environment_patches()
+from meerschaum.config.environment import (
+    apply_environment_patches as _apply_environment_patches,
+    apply_environment_uris as _apply_environment_uris,
+)
+_apply_environment_uris()
+_apply_environment_patches()
 
 
 from meerschaum.config._paths import PATCH_DIR_PATH, PERMANENT_PATCH_DIR_PATH
