@@ -119,8 +119,6 @@ default_system_config = {
         'api': {
         },
     },
-    'api': 'MRSM{api}',
-    'webterm': 'MRSM{api:webterm}',
     'cli': {
         'max_daemons': (multiprocessing.cpu_count() * 3),
         'refresh_seconds': 0.1,
@@ -153,6 +151,7 @@ default_system_config = {
             'daemon',
             'sql',
             'register plugin',
+            'copy ',
         ],
     },
     'experimental': {
@@ -164,7 +163,7 @@ default_system_config = {
         'uv_pip': True,
         'systemd_healthcheck': False,
         'valkey_session_cache': True,
-        'cli_daemon': True,
+        'cli_daemon': False,
     },
 }
 
