@@ -13,14 +13,14 @@ from meerschaum.utils.packages import attempt_import
 
 @classmethod
 def from_uri(
-        cls,
-        uri: str,
-        label: Optional[str] = None,
-        as_dict: bool = False,
-    ) -> Union[
-        'meerschaum.connectors.SQLConnector',
-        Dict[str, Union[str, int]],
-    ]:
+    cls,
+    uri: str,
+    label: Optional[str] = None,
+    as_dict: bool = False,
+) -> Union[
+    'meerschaum.connectors.SQLConnector',
+    Dict[str, Union[str, int]],
+]:
     """
     Create a new SQLConnector from a URI string.
 
@@ -51,7 +51,7 @@ def from_uri(
     if 'database' not in params:
         error("Unable to determine the database from the provided URI.")
 
-    if flavor in ('sqlite', 'duckdb'):
+    if flavor in ('sqlite', 'duckdb', 'geopackage'):
         if params['database'] == ':memory:':
             params['label'] = label or f'memory_{flavor}'
         else:
@@ -97,7 +97,7 @@ def parse_uri(uri: str) -> Dict[str, Any]:
     >>> 
     """
     from urllib.parse import parse_qs, urlparse
-    sqlalchemy = attempt_import('sqlalchemy')
+    sqlalchemy = attempt_import('sqlalchemy', lazy=False)
     parser = sqlalchemy.engine.url.make_url
     params = parser(uri).translate_connect_args()
     params['flavor'] = uri.split(':')[0].split('+')[0]
@@ -107,4 +107,7 @@ def parse_uri(uri: str) -> Dict[str, Any]:
         parsed_uri = urlparse(uri)
         for key, value in parse_qs(parsed_uri.query).items():
             params.update({key: value[0]})
+
+        if '--search_path' in params.get('options', ''):
+            params.update({'schema': params['options'].replace('--search_path=', '', 1)})
     return params
