@@ -7,19 +7,54 @@ Manage users via the SQL Connector
 """
 
 from __future__ import annotations
+
+import meerschaum as mrsm
 from meerschaum.utils.typing import SuccessTuple, Optional, Any, Dict, List, Union
 
+
+def get_users_pipe(self) -> mrsm.Pipe:
+    """
+    Return the internal metadata pipe for users management.
+    """
+    if '_users_pipe' in self.__dict__:
+        return self._users_pipe
+
+    cache_connector = self.__dict__.get('_cache_connector', None)
+    self._users_pipe = mrsm.Pipe(
+        'mrsm', 'users',
+        temporary=True,
+        cache=True,
+        cache_connector_keys=cache_connector,
+        static=True,
+        null_indices=False,
+        enforce=False,
+        autoincrement=True,
+        columns={
+            'primary': 'user_id',
+        },
+        dtypes={
+            'user_id': 'int',
+            'username': 'string',
+            'attributes': 'json',
+            'user_type': 'string',
+        },
+        indices={
+            'unique': 'username',
+        },
+    )
+    return self._users_pipe
+
+
 def register_user(
-        self,
-        user: meerschaum.core.User,
-        debug: bool = False,
-        **kw: Any
-    ) -> SuccessTuple:
+    self,
+    user: mrsm.core.User,
+    debug: bool = False,
+    **kw: Any
+) -> SuccessTuple:
     """Register a new user."""
-    from meerschaum.utils.warnings import warn, error, info
     from meerschaum.utils.packages import attempt_import
     from meerschaum.utils.sql import json_flavors
-    sqlalchemy = attempt_import('sqlalchemy')
+    sqlalchemy = attempt_import('sqlalchemy', lazy=False)
 
     valid_tuple = valid_username(user.username)
     if not valid_tuple[0]:
@@ -36,12 +71,14 @@ def register_user(
 
     import json
     bind_variables = {
-        'username' : user.username,
-        'email' : user.email,
-        'password_hash' : user.password_hash,
-        'user_type' : user.type,
-        'attributes' : (
-            json.dumps(user.attributes) if self.flavor not in json_flavors else user.attributes
+        'username': user.username,
+        'email': user.email,
+        'password_hash': user.password_hash,
+        'user_type': user.type,
+        'attributes': (
+            json.dumps(user.attributes)
+            if self.flavor not in json_flavors
+            else user.attributes
         ),
     }
     if old_id is not None:
@@ -57,9 +94,10 @@ def register_user(
         return False, f"Failed to register user '{user}'."
     return True, f"Successfully registered user '{user}'."
 
+
 def valid_username(username: str) -> SuccessTuple:
     """Verify that a given username is valid."""
-    from meerschaum.config.static import STATIC_CONFIG
+    from meerschaum._internal.static import STATIC_CONFIG
     fail_reasons = []
 
     min_length = STATIC_CONFIG['users']['min_username_length']
@@ -70,12 +108,12 @@ def valid_username(username: str) -> SuccessTuple:
     if len(username) > max_length:
         fail_reasons.append(f"Usernames must contain {max_length} or fewer characters.")
 
-    acceptable_chars = {'_', '-'}
+    acceptable_chars = {'_', '-', '.', '@'}
     for c in username:
         if not c.isalnum() and c not in acceptable_chars:
             fail_reasons.append(
                 (
-                    f"Usernames may only contain alphanumeric characters " +
+                    "Usernames may only contain alphanumeric characters " +
                     "and the following special characters: "
                     + str(list(acceptable_chars))
                 )
@@ -92,22 +130,22 @@ def valid_username(username: str) -> SuccessTuple:
 
 
 def edit_user(
-        self,
-        user: meerschaum.core.User,
-        debug: bool = False,
-        **kw: Any
-    ) -> SuccessTuple:
+    self,
+    user: 'mrsm.core.User',
+    debug: bool = False,
+    **kw: Any
+) -> SuccessTuple:
     """Update an existing user's metadata."""
     from meerschaum.utils.packages import attempt_import
-    sqlalchemy = attempt_import('sqlalchemy')
-    from meerschaum.connectors.sql.tables import get_tables
     from meerschaum.utils.sql import json_flavors
+    sqlalchemy = attempt_import('sqlalchemy', lazy=False)
+    from meerschaum.connectors.sql.tables import get_tables
     users_tbl = get_tables(mrsm_instance=self, debug=debug)['users']
 
     user_id = user.user_id if user.user_id is not None else self.get_user_id(user, debug=debug)
     if user_id is None:
         return False, (
-            f"User '{user.username}' does not exist. " +
+            f"User '{user.username}' does not exist. "
             f"Register user '{user.username}' before editing."
         )
     user.user_id = user_id
@@ -127,7 +165,7 @@ def edit_user(
         bind_variables['email'] = user.email
     if user.attributes is not None and user.attributes != {}:
         bind_variables['attributes'] = (
-            json.dumps(user.attributes) if self.flavor in ('duckdb',)
+            json.dumps(user.attributes) if self.flavor not in json_flavors
             else user.attributes
         )
     if user.type != '':
@@ -145,15 +183,16 @@ def edit_user(
         return False, f"Failed to edit user '{user}'."
     return True, f"Successfully edited user '{user}'."
 
+
 def get_user_id(
-        self,
-        user: meerschaum.core.User,
-        debug : bool = False
-    ) -> Optional[int]:
+    self,
+    user: 'mrsm.core.User',
+    debug: bool = False
+) -> Optional[int]:
     """If a user is registered, return the `user_id`."""
     ### ensure users table exists
     from meerschaum.utils.packages import attempt_import
-    sqlalchemy = attempt_import('sqlalchemy')
+    sqlalchemy = attempt_import('sqlalchemy', lazy=False)
     from meerschaum.connectors.sql.tables import get_tables
     users_tbl = get_tables(mrsm_instance=self, debug=debug)['users']
 
@@ -168,17 +207,17 @@ def get_user_id(
     return None
 
 def get_user_attributes(
-        self,
-        user: meerschaum.core.User,
-        debug: bool = False
-    ) -> Union[Dict[str, Any], None]:
+    self,
+    user: 'mrsm.core.User',
+    debug: bool = False
+) -> Union[Dict[str, Any], None]:
     """
     Return the user's attributes.
     """
     ### ensure users table exists
     from meerschaum.utils.warnings import warn
     from meerschaum.utils.packages import attempt_import
-    sqlalchemy = attempt_import('sqlalchemy')
+    sqlalchemy = attempt_import('sqlalchemy', lazy=False)
     from meerschaum.connectors.sql.tables import get_tables
     users_tbl = get_tables(mrsm_instance=self, debug=debug)['users']
 
@@ -194,31 +233,31 @@ def get_user_attributes(
         try:
             result = dict(result)
             _parsed = True
-        except Exception as e:
+        except Exception:
             _parsed = False
         if not _parsed:
             try:
                 import json
                 result = json.loads(result)
                 _parsed = True
-            except Exception as e:
+            except Exception:
                 _parsed = False
         if not _parsed:
             warn(f"Received unexpected type for attributes: {result}")
     return result
 
 def delete_user(
-        self,
-        user: meerschaum.core.User,
-        debug: bool = False
-    ) -> SuccessTuple:
+    self,
+    user: 'mrsm.core.User',
+    debug: bool = False
+) -> SuccessTuple:
     """Delete a user's record from the users table."""
     ### ensure users table exists
     from meerschaum.connectors.sql.tables import get_tables
     users_tbl = get_tables(mrsm_instance=self, debug=debug)['users']
     plugins = get_tables(mrsm_instance=self, debug=debug)['plugins']
     from meerschaum.utils.packages import attempt_import
-    sqlalchemy = attempt_import('sqlalchemy')
+    sqlalchemy = attempt_import('sqlalchemy', lazy=False)
 
     user_id = user.user_id if user.user_id is not None else self.get_user_id(user, debug=debug)
 
@@ -238,11 +277,12 @@ def delete_user(
 
     return True, f"Successfully deleted user '{user}'"
 
+
 def get_users(
-        self,
-        debug: bool = False,
-        **kw: Any
-    ) -> List[str]:
+    self,
+    debug: bool = False,
+    **kw: Any
+) -> List[str]:
     """
     Get the registered usernames.
     """
@@ -250,18 +290,19 @@ def get_users(
     from meerschaum.connectors.sql.tables import get_tables
     users_tbl = get_tables(mrsm_instance=self, debug=debug)['users']
     from meerschaum.utils.packages import attempt_import
-    sqlalchemy = attempt_import('sqlalchemy')
+    sqlalchemy = attempt_import('sqlalchemy', lazy=False)
 
     query = sqlalchemy.select(users_tbl.c.username)
 
     return list(self.read(query, debug=debug)['username'])
 
+
 def get_user_password_hash(
-        self,
-        user: meerschaum.core.User,
-        debug: bool = False,
-        **kw: Any
-    ) -> Optional[str]:
+    self,
+    user: 'mrsm.core.User',
+    debug: bool = False,
+    **kw: Any
+) -> Optional[str]:
     """
     Return the password has for a user.
     **NOTE**: This may be dangerous and is only allowed if the security settings explicity allow it.
@@ -270,7 +311,7 @@ def get_user_password_hash(
     from meerschaum.connectors.sql.tables import get_tables
     users_tbl = get_tables(mrsm_instance=self, debug=debug)['users']
     from meerschaum.utils.packages import attempt_import
-    sqlalchemy = attempt_import('sqlalchemy')
+    sqlalchemy = attempt_import('sqlalchemy', lazy=False)
 
     if user.user_id is not None:
         user_id = user.user_id
@@ -278,7 +319,7 @@ def get_user_password_hash(
             dprint(f"Already given user_id: {user_id}")
     else:
         if debug:
-            dprint(f"Fetching user_id...")
+            dprint("Fetching user_id...")
         user_id = self.get_user_id(user, debug=debug)
 
     if user_id is None:
@@ -288,19 +329,20 @@ def get_user_password_hash(
 
     return self.value(query, debug=debug)
 
+
 def get_user_type(
-        self,
-        user: meerschaum.core.User,
-        debug: bool = False,
-        **kw: Any
-    ) -> Optional[str]:
+    self,
+    user: 'mrsm.core.User',
+    debug: bool = False,
+    **kw: Any
+) -> Optional[str]:
     """
     Return the user's type.
     """
     from meerschaum.connectors.sql.tables import get_tables
     users_tbl = get_tables(mrsm_instance=self, debug=debug)['users']
     from meerschaum.utils.packages import attempt_import
-    sqlalchemy = attempt_import('sqlalchemy')
+    sqlalchemy = attempt_import('sqlalchemy', lazy=False)
 
     user_id = user.user_id if user.user_id is not None else self.get_user_id(user, debug=debug)
 
