@@ -7,7 +7,11 @@ Formatting functions for printing pipes
 """
 
 from __future__ import annotations
-from meerschaum.utils.typing import PipesDict, Dict, Union, Optional
+import json
+import meerschaum as mrsm
+from meerschaum.utils.typing import PipesDict, Dict, Union, Optional, SuccessTuple, Any, List
+from meerschaum.config import get_config
+
 
 def pprint_pipes(pipes: PipesDict) -> None:
     """Print a stylized tree of a Pipes dictionary.
@@ -16,7 +20,6 @@ def pprint_pipes(pipes: PipesDict) -> None:
     from meerschaum.utils.packages import attempt_import, import_rich
     from meerschaum.utils.misc import sorted_dict, replace_pipes_in_dict
     from meerschaum.utils.formatting import UNICODE, ANSI, CHARSET, pprint, colored, get_console
-    from meerschaum.config import get_config
     import copy
     rich = import_rich('rich', warn=False)
     Text = None
@@ -200,15 +203,15 @@ def pprint_pipes(pipes: PipesDict) -> None:
     columns = Columns(cols)
     get_console().print(columns)
 
+
 def pprint_pipe_columns(
         pipe: meerschaum.Pipe,
         nopretty: bool = False,
         debug: bool = False,
     ) -> None:
     """Pretty-print a pipe's columns."""
-    import json
     from meerschaum.utils.warnings import info
-    from meerschaum.utils.formatting import pprint, print_tuple, get_console
+    from meerschaum.utils.formatting import UNICODE, ANSI, pprint, print_tuple, get_console
     from meerschaum.utils.formatting._shell import make_header
     from meerschaum.utils.packages import attempt_import, import_rich
 
@@ -224,27 +227,48 @@ def pprint_pipe_columns(
 
     def _pretty_print():
         rich = import_rich()
-        rich_table = attempt_import('rich.table')
+        rich_table, rich_text, rich_box = attempt_import(
+            'rich.table',
+            'rich.text',
+            'rich.box',
+        )
 
-        table = rich_table.Table(title=f"Data Types for {pipe}")
-        table.add_column('Column')
-        table.add_column('DB Type', justify='right')
-        table.add_column('PD Type', justify='left')
+        table = rich_table.Table(box=(rich_box.MINIMAL if UNICODE else rich_box.ASCII))
+        styles = {
+            'column': ('white' if ANSI else None),
+            'db_type': ('cyan' if ANSI else None),
+            'pd_type': ('yellow' if ANSI else None),
+        }
+        table.add_column(
+            rich_text.Text('Column', style=styles['column'])
+        )
+        table.add_column(
+            rich_text.Text('DB Type', style=styles['db_type']),
+            justify = 'right',
+        )
+        table.add_column(
+            rich_text.Text('PD Type', style=styles['pd_type']),
+            justify = 'left',
+        )
 
         info(make_header(f"\nIndex Columns for {pipe}:"), icon=False)
         if _cols:
             pprint(_cols, nopretty=nopretty)
             print()
         else:
-            print_tuple((False, f"No registered columns for {pipe}."))
+            print_tuple((False, f"No registered columns for {pipe}."), calm=True)
 
         for c, t in pipe.dtypes.items():
-            table.add_row(c, _cols_types.get(c, None), t)
+            table.add_row(
+                rich_text.Text(c, style=styles['column']),
+                rich_text.Text(_cols_types.get(c, 'None'), style=styles['db_type']),
+                rich_text.Text(t, style=styles['pd_type']),
+            )
 
         if pipe.dtypes:
             get_console().print(table)
         else:
-            print_tuple((False, f"No table columns for {pipe}. Does the pipe exist?"))
+            print_tuple((False, f"No table columns for {pipe}. Does the pipe exist?"), calm=True)
 
     if nopretty:
         _nopretty_print()
@@ -253,19 +277,31 @@ def pprint_pipe_columns(
 
 
 def pipe_repr(
-        pipe: 'meerschaum.Pipe',
-        as_rich_text: bool=False,
-        ansi: Optional[bool] = None,
-    ) -> Union[str, 'rich.text.Text']:
+    pipe: Union[mrsm.Pipe, Dict[str, Any]],
+    as_rich_text: bool = False,
+    ansi: Optional[bool] = None,
+) -> Union[str, 'rich.text.Text']:
     """
     Return a formatted string for representing a `meerschaum.Pipe`.
     """
-    from meerschaum.config import get_config
-    from meerschaum.utils.formatting import UNICODE, ANSI, CHARSET, colored, rich_text_to_str
+    from meerschaum.utils.formatting import ANSI, colored, rich_text_to_str
     from meerschaum.utils.packages import import_rich, attempt_import
-    rich = import_rich()
+    import meerschaum as mrsm
+
+    _ = import_rich()
     Text = attempt_import('rich.text').Text
 
+    if isinstance(pipe, mrsm.Pipe):
+        connector_keys = pipe.connector_keys
+        metric_key = pipe.metric_key
+        location_key = pipe.location_key
+        instance_keys = pipe.instance_keys
+    else:
+        connector_keys = pipe.get('connector_keys')
+        metric_key = pipe.get('metric_key')
+        location_key = pipe.get('location_key')
+        instance_keys = pipe.get('instance_keys', get_config('meerschaum', 'instance'))
+
     styles = get_config('formatting', 'pipes', '__repr__', 'ansi', 'styles')
     if not ANSI or (ansi is False):
         styles = {k: '' for k in styles}
@@ -275,56 +311,247 @@ def pipe_repr(
     )
     text_obj = (
         Text.from_markup(_pipe_style_prefix + "Pipe(" + _pipe_style_suffix)
-        + colored(("'" + pipe.connector_keys + "'"), style=styles['connector'], as_rich_text=True)
+        + colored(("'" + connector_keys + "'"), style=styles['connector'], as_rich_text=True)
         + Text.from_markup(_pipe_style_prefix + ", " + _pipe_style_suffix)
-        + colored(("'" + pipe.metric_key + "'"), style=styles['metric'], as_rich_text=True)
+        + colored(("'" + metric_key + "'"), style=styles['metric'], as_rich_text=True)
         + (
             (
                 colored(', ', style=styles['punctuation'], as_rich_text=True)
                 + colored(
-                    ("'" + pipe.location_key + "'"),
+                    ("'" + location_key + "'"),
                     style=styles['location'], as_rich_text=True
                 )
-            ) if pipe.location_key is not None
+            ) if location_key is not None
             else colored('', style='', as_rich_text=True)
         ) + (
             ( ### Add the `instance=` argument.
                 colored(', instance=', style=styles['punctuation'], as_rich_text=True)
                 + colored(
-                    ("'" + pipe.instance_keys + "'"),
+                    ("'" + instance_keys + "'"),
                     style=styles['instance'], as_rich_text=True
                 )
-            ) if pipe.instance_keys != get_config('meerschaum', 'instance')
+            ) if instance_keys != get_config('meerschaum', 'instance')
             else colored('', style='', as_rich_text=True)
         )
         + Text.from_markup(_pipe_style_prefix + ")" + _pipe_style_suffix)
     )
     if as_rich_text:
         return text_obj
-    return rich_text_to_str(text_obj)
+    return rich_text_to_str(text_obj).replace('\n', '')
+
 
 
 def highlight_pipes(message: str) -> str:
     """
     Add syntax highlighting to an info message containing stringified `meerschaum.Pipe` objects.
     """
-    if 'Pipe(' not in message or ')' not in message:
+    if 'Pipe(' not in message:
         return message
-    from meerschaum import Pipe
+
+    from meerschaum.utils.misc import parse_arguments_str
     segments = message.split('Pipe(')
     msg = ''
-    _d = {}
     for i, segment in enumerate(segments):
-        if ',' in segment and ')' in segment:
-            paren_index = segment.find(')') + 1
-            code = "_d['pipe'] = Pipe(" + segment[:paren_index]
-            try:
-                exec(code)
-                _to_add = pipe_repr(_d['pipe']) + segment[paren_index:]
-            except Exception as e:
-                _to_add = 'Pipe(' + segment
-            msg += _to_add
+        if i == 0:
+            msg += segment
             continue
-        msg += segment
+
+        paren_index = segment.find(')')
+        if paren_index == -1:
+            msg += 'Pipe(' + segment
+            continue
+        
+        pipe_args_str = segment[:paren_index]
+        try:
+            args, kwargs = parse_arguments_str(pipe_args_str)
+            pipe_dict = {
+                'connector_keys': args[0],
+                'metric_key': args[1],
+            }
+            if len(args) > 2:
+                pipe_dict['location_key'] = args[2]
+            if 'instance' in kwargs:
+                pipe_dict['instance_keys'] = kwargs['instance']
+            
+            _to_add = pipe_repr(pipe_dict) + segment[paren_index + 1:]
+        except Exception:
+            _to_add = 'Pipe(' + segment
+        msg += _to_add
     return msg
 
+
+def format_pipe_success_tuple(
+    pipe: mrsm.Pipe,
+    success_tuple: SuccessTuple,
+    nopretty: bool = False,
+) -> str:
+    """
+    Return a formatted string of a pipe and its resulting SuccessTuple.
+
+    Parameters
+    ----------
+    pipe: mrsm.Pipe
+        The pipe to print.
+
+    success_tuple: SuccessTuple
+        The output of a pipe's actions (e.g. `pipe.sync()`).
+
+    nopretty: bool, default False
+        If `True`, return a JSON-formatted string instead.
+
+    Returns
+    -------
+    The formatted string of the pipe and its resulting SuccessTuple.
+        
+    """
+    from meerschaum.utils.formatting import UNICODE
+    underline = '\u2015' if UNICODE else '-'
+    success, message = success_tuple
+    success_prefix = (get_config('formatting', 'emoji', 'success_calm') + ' ') if UNICODE else ''
+    failure_prefix = (get_config('formatting', 'emoji', 'failure_calm') + ' ') if UNICODE else ''
+    pipe_str = (success_prefix if success else failure_prefix) + str(pipe)
+    return (
+        (
+            pipe_str
+            + '\n'
+            ### The extra +1 at the end is for double-width emoji.
+            + (underline * (len(pipe_str) + (1 if UNICODE else 0)))
+            + '\n'
+            + str(message)
+            + '\n\n'
+        ) if not nopretty else (
+            json.dumps({
+                'pipe': pipe.meta,
+                'success': success,
+                'message': message,
+            })
+        )
+    )
+
+
+def print_pipes_results(
+        pipes_results: Dict[mrsm.Pipe, SuccessTuple],
+        success_header: Optional[str] = 'Successes',
+        failure_header: Optional[str] = 'Failures',
+        nopretty: bool = False,
+        **kwargs: Any
+    ) -> None:
+    """
+    Print the pipes and their result SuccessTuples.
+
+    Parameters
+    ----------
+    pipes_results: Dict[mrsm.Pipe, SuccessTuple]
+        A dictionary mapping pipes to their resulting SuccessTuples.
+
+    success_header: Optional[str], default 'Successes'
+        The header to print above the successful pipes.
+
+    failure_header: Optional[str], default 'Fails'
+        The header to print above the failed pipes.
+
+    kwargs: Any
+        All other keyword arguments are passed to `meerschaum.utils.misc.print_options`.
+    """
+    from meerschaum.utils.misc import print_options
+    successes = [pipe for pipe, (success, msg) in pipes_results.items() if success]
+    fails = [pipe for pipe, (success, msg) in pipes_results.items() if success]
+    success_options = [
+        format_pipe_success_tuple(pipe, success_tuple, nopretty=nopretty)
+        for pipe, success_tuple in pipes_results.items()
+        if success_tuple[0]
+    ]
+    failure_options = [
+        format_pipe_success_tuple(pipe, success_tuple, nopretty=nopretty)
+        for pipe, success_tuple in pipes_results.items()
+        if not success_tuple[0]
+    ]
+
+    if success_options:
+        print_options(
+            success_options,
+            header = success_header,
+            nopretty = nopretty,
+            **kwargs
+        )
+    if failure_options:
+        print_options(
+            failure_options,
+            header = failure_header,
+            nopretty = nopretty,
+            **kwargs
+        )
+
+
+def extract_stats_from_message(
+    message: str,
+    stat_keys: Optional[List[str]] = None,
+) -> Dict[str, int]:
+    """
+    Given a sync message, return the insert, update, upsert stats from within.
+
+    Parameters
+    ----------
+    message: str
+        The message to parse for statistics.
+
+    stat_keys: Optional[List[str]], default None
+        If provided, search for these words (case insensitive) in the message.
+        Defaults to `['inserted', 'updated', 'upserted']`.
+
+    Returns
+    -------
+    A dictionary mapping the stat keys to the total number of rows affected.
+    """
+    stat_keys = stat_keys or ['inserted', 'updated', 'upserted', 'checked']
+    lines_stats = [extract_stats_from_line(line, stat_keys) for line in message.split('\n')]
+    message_stats = {
+        stat_key: sum(stats.get(stat_key, 0) for stats in lines_stats)
+        for stat_key in stat_keys
+    }
+    return message_stats
+
+
+def extract_stats_from_line(
+    line: str,
+    stat_keys: List[str],
+) -> Dict[str, int]:
+    """
+    Return the insert, update, upsert stats from a single line.
+    """
+    stats = {key: 0 for key in stat_keys}
+
+    for stat_key in stat_keys:
+        search_key = stat_key.lower()
+        if search_key not in line.lower():
+            continue
+
+        ### the count may be formatted with commas
+        line = line.replace(',', '')
+
+        ### stat_text starts with the digits we want.
+        try:
+            stat_text = line.lower().split(search_key + ' ')[1]
+        except IndexError:
+            continue
+
+        ### find the first non-digit value.
+        end_of_num_ix = -1
+        for i, char in enumerate(stat_text):
+            if not char.isdigit():
+                end_of_num_ix = i
+                break
+            if i == len(stat_text) - 1:
+                end_of_num_ix = i + 1
+        if end_of_num_ix == -1:
+            continue
+
+        try:
+            stat_number = int(stat_text[:end_of_num_ix])
+        except Exception as e:
+            print(e)
+            continue
+
+        stats[search_key] += stat_number
+
+    return stats
