@@ -36,6 +36,7 @@ exists_queries = {
 version_queries = {
     'default': "SELECT VERSION() AS {version_name}",
     'sqlite': "SELECT SQLITE_VERSION() AS {version_name}",
+    'geopackage': "SELECT SQLITE_VERSION() AS {version_name}",
     'mssql': "SELECT @@version",
     'oracle': "SELECT version from PRODUCT_COMPONENT_VERSION WHERE rownum = 1",
 }
@@ -50,6 +51,7 @@ DROP_IF_EXISTS_FLAVORS = {
     'mysql',
     'mariadb',
     'sqlite',
+    'geopackage',
 }
 DROP_INDEX_IF_EXISTS_FLAVORS = {
     'mssql',
@@ -58,6 +60,7 @@ DROP_INDEX_IF_EXISTS_FLAVORS = {
     'postgresql',
     'postgis',
     'sqlite',
+    'geopackage',
     'citus',
 }
 SKIP_AUTO_INCREMENT_FLAVORS = {'citus', 'duckdb'}
@@ -223,6 +226,13 @@ UPDATE_QUERIES = {
         SELECT {patch_cols_str} FROM {patch_table_name} AS p
         """,
     ],
+    'geopackage-upsert': """
+        INSERT INTO {target_table_name} ({patch_cols_str})
+        SELECT {patch_cols_str}
+        FROM {patch_table_name}
+        WHERE true
+        ON CONFLICT ({join_cols_str}) DO {update_or_nothing} {sets_subquery_none_excluded}
+    """,
 }
 columns_types_queries = {
     'default': """
@@ -250,6 +260,19 @@ columns_types_queries = {
         WHERE m.type = 'table'
             AND m.name IN ('{table}', '{table_trunc}')
     """,
+    'geopackage': """
+        SELECT
+            '' "database",
+            '' "schema",
+            m.name "table",
+            p.name "column",
+            p.type "type"
+        FROM sqlite_master m
+        LEFT OUTER JOIN pragma_table_info(m.name) p
+            ON m.name <> p.name
+        WHERE m.type = 'table'
+            AND m.name IN ('{table}', '{table_trunc}')
+    """,
     'mssql': """
         SELECT
             TABLE_CATALOG AS [database],
@@ -388,6 +411,49 @@ columns_indices_queries = {
             index_type
         FROM primary_key_columns
     """,
+    'geopackage': """
+        WITH indexed_columns AS (
+            SELECT
+                '{table}' AS table_name,
+                pi.name AS column_name,
+                i.name AS index_name,
+                'INDEX' AS index_type
+            FROM
+                sqlite_master AS i,
+                pragma_index_info(i.name) AS pi
+            WHERE
+                i.type = 'index'
+                AND i.tbl_name = '{table}'
+        ),
+        primary_key_columns AS (
+            SELECT
+                '{table}' AS table_name,
+                ti.name AS column_name,
+                'PRIMARY_KEY' AS index_name,
+                'PRIMARY KEY' AS index_type
+            FROM
+                pragma_table_info('{table}') AS ti
+            WHERE
+                ti.pk > 0
+        )
+        SELECT
+            NULL AS "database",
+            NULL AS "schema",
+            "table_name" AS "table",
+            "column_name" AS "column",
+            "index_name" AS "index",
+            "index_type"
+        FROM indexed_columns
+        UNION ALL
+        SELECT
+            NULL AS "database",
+            NULL AS "schema",
+            table_name AS "table",
+            column_name AS "column",
+            index_name AS "index",
+            index_type
+        FROM primary_key_columns
+    """,
     'mssql': """
         SELECT
             NULL AS [database],
@@ -504,6 +570,11 @@ reset_autoincrement_queries: Dict[str, Union[str, List[str]]] = {
         SET seq = {val}
         WHERE name = '{table}'
     """,
+    'geopackage': """
+        UPDATE sqlite_sequence
+        SET seq = {val}
+        WHERE name = '{table}'
+    """,
     'oracle': (
         "ALTER TABLE {table_name} MODIFY {column_name} "
         "GENERATED BY DEFAULT ON NULL AS IDENTITY (START WITH {val_plus_1})"
@@ -518,6 +589,7 @@ table_wrappers = {
     'postgresql'    : ('"', '"'),
     'postgis'       : ('"', '"'),
     'sqlite'        : ('"', '"'),
+    'geopackage'    : ('"', '"'),
     'mysql'         : ('`', '`'),
     'mariadb'       : ('`', '`'),
     'mssql'         : ('[', ']'),
@@ -534,7 +606,8 @@ max_name_lens = {
     'timescaledb-ha': 64,
     'citus'         : 64,
     'cockroachdb'   : 64,
-    'sqlite'        : 1024, ### Probably more, but 1024 seems more than reasonable.
+    'sqlite'        : 1024,
+    'geopackage'    : 1024,
     'mysql'         : 64,
     'mariadb'       : 64,
 }
@@ -549,6 +622,7 @@ json_flavors = {
 NO_SCHEMA_FLAVORS = {
     'oracle',
     'sqlite',
+    'geopackage',
     'mysql',
     'mariadb',
     'duckdb',
@@ -573,6 +647,7 @@ OMIT_NULLSFIRST_FLAVORS = {
 SINGLE_ALTER_TABLE_FLAVORS = {
     'duckdb',
     'sqlite',
+    'geopackage',
     'mssql',
     'oracle',
 }
@@ -582,6 +657,7 @@ NO_CTE_FLAVORS = {
 }
 NO_SELECT_INTO_FLAVORS = {
     'sqlite',
+    'geopackage',
     'oracle',
     'mysql',
     'mariadb',
@@ -629,6 +705,7 @@ def dateadd_str(
         - `'mysql'`
         - `'mariadb'`
         - `'sqlite'`
+        - `'geopackage'`
         - `'oracle'`
 
     datepart: str, default `'day'`
@@ -763,7 +840,7 @@ def dateadd_str(
         )
         da = (f"DATE_ADD({begin}, INTERVAL {number} {datepart})" if number != 0 else begin)
 
-    elif flavor == 'sqlite':
+    elif flavor in ('sqlite', 'geopackage'):
         da = f"datetime({begin}, '{number} {datepart}')"
 
     elif flavor == 'oracle':
@@ -909,7 +986,7 @@ def sql_item_name(item: str, flavor: str, schema: Optional[str] = None) -> str:
         wrappers = table_wrappers.get(flavor, table_wrappers['default'])
 
     ### NOTE: SQLite does not support schemas.
-    if flavor == 'sqlite':
+    if flavor in ('sqlite', 'geopackage'):
         schema = None
     elif flavor == 'mssql' and str(item).startswith('#'):
         schema = None
@@ -1293,7 +1370,7 @@ def get_table_cols_types(
         schema = None
     if schema is None:
         schema = DEFAULT_SCHEMA_FLAVORS.get(flavor, None)
-    if flavor in ('sqlite', 'duckdb', 'oracle'):
+    if flavor in ('sqlite', 'duckdb', 'oracle', 'geopackage'):
         database = None
     table_trunc = truncate_item_name(table, flavor=flavor)
     table_lower = table.lower()
@@ -1472,7 +1549,7 @@ def get_table_cols_indices(
         schema = None
     if schema is None:
         schema = DEFAULT_SCHEMA_FLAVORS.get(flavor, None)
-    if flavor in ('sqlite', 'duckdb', 'oracle'):
+    if flavor in ('sqlite', 'duckdb', 'oracle', 'geopackage'):
         database = None
     table_trunc = truncate_item_name(table, flavor=flavor)
     table_lower = table.lower()
@@ -1662,7 +1739,7 @@ def get_update_queries(
     if not flavor:
         raise ValueError("Provide a flavor if using a SQLAlchemy session.")
     if (
-        flavor == 'sqlite'
+        flavor in ('sqlite', 'geopackage')
         and isinstance(connectable, SQLConnector)
         and connectable.db_version < '3.33.0'
     ):
@@ -1776,7 +1853,7 @@ def get_update_queries(
                         if c_name in utc_value_cols
                         else ''
                     ))
-                    if flavor != 'sqlite'
+                    if flavor not in ('sqlite', 'geopackage')
                     else ('', '', '')
                 )
             )
@@ -2069,6 +2146,7 @@ def get_create_table_queries(
     primary_key_db_type: Optional[str] = None,
     autoincrement: bool = False,
     datetime_column: Optional[str] = None,
+    _parse_dtypes: bool = True,
 ) -> List[str]:
     """
     Return a query to create a new table from a `SELECT` query or a `dtypes` dictionary.
@@ -2102,6 +2180,10 @@ def get_create_table_queries(
         If provided, include this column in the primary key.
         Applicable to TimescaleDB only.
 
+    _parse_dtypes: bool, default True
+        If `True`, cast Pandas dtypes to SQL dtypes.
+        Otherwise pass through the given value directly.
+
     Returns
     -------
     A `CREATE TABLE` (or `SELECT INTO`) query for the database flavor.
@@ -2123,6 +2205,7 @@ def get_create_table_queries(
         primary_key_db_type=primary_key_db_type,
         autoincrement=(autoincrement and flavor not in SKIP_AUTO_INCREMENT_FLAVORS),
         datetime_column=datetime_column,
+        _parse_dtypes=_parse_dtypes,
     )
 
 
@@ -2135,6 +2218,7 @@ def _get_create_table_query_from_dtypes(
     primary_key_db_type: Optional[str] = None,
     autoincrement: bool = False,
     datetime_column: Optional[str] = None,
+    _parse_dtypes: bool = True,
 ) -> List[str]:
     """
     Create a new table from a `dtypes` dictionary.
@@ -2151,12 +2235,22 @@ def _get_create_table_query_from_dtypes(
             (
                 primary_key,
                 get_db_type_from_pd_type(dtypes.get(primary_key, 'int') or 'int', flavor=flavor)
+            ) if _parse_dtypes else (
+                primary_key,
+                dtypes.get(primary_key, 'INT') or 'INT'
             )
         ]
         if primary_key
         else []
     ) + [
-        (col, get_db_type_from_pd_type(typ, flavor=flavor))
+        (
+            col,
+            (
+                get_db_type_from_pd_type(typ, flavor=flavor)
+                if _parse_dtypes
+                else typ
+            )
+        )
         for col, typ in dtypes.items()
         if col != primary_key
     ]
@@ -2183,7 +2277,7 @@ def _get_create_table_query_from_dtypes(
         )) if autoincrement or primary_key not in dtypes else ''
         col_name = sql_item_name(primary_key, flavor=flavor, schema=None)
 
-        if flavor == 'sqlite':
+        if flavor in ('sqlite', 'geopackage'):
             query += (
                 f"\n    {col_name} "
                 + (f"{col_db_type}" if not auto_increment_str else 'INTEGER')
@@ -2234,6 +2328,7 @@ def _get_create_table_query_from_cte(
     primary_key_db_type: Optional[str] = None,
     autoincrement: bool = False,
     datetime_column: Optional[str] = None,
+    _parse_dtypes=None,
 ) -> List[str]:
     """
     Create a new table from a CTE query.
@@ -2315,7 +2410,7 @@ def _get_create_table_query_from_cte(
                 f"ADD PRIMARY KEY ({primary_key_name})"
             ),
         ]
-    elif flavor in ('sqlite', 'mysql', 'mariadb', 'duckdb', 'oracle'):
+    elif flavor in ('sqlite', 'mysql', 'mariadb', 'duckdb', 'oracle', 'geopackage'):
         create_table_queries = [
             (
                 f"CREATE TABLE {new_table_name} AS\n"
